#ifndef(FINDAFFINE_INC_TEMP)
#declare FINDAFFINE_INC_TEMP = version;
#version unofficial megapov 1.21;

#ifdef(View_POV_Include_Stack)
   #debug "including FindAffine.inc\n"
#end      



#include "debug.inc"

#macro swap( a, b )
  #local temp = a;
  #declare a = b;
  #declare b = temp;
#end     

// finds the inverse of A using GaussJordan elimination
// and stores it in I...only for 4x4 matrices
// B will contain the solution vectors
#macro GaussJordan3x3( A, B, I )   
  Debug_Message( "Entering GaussJordan4x4\n" )

  // arrays for pivoting book-keeping
  #local indxc = array[ 3 ];
  #local indxr = array[ 3 ];
  #local ipiv = array[ 3 ];
  
  #local j = 0;
  #while( j < 3 )
    #local ipiv[ j ] = 0; 
        
    #local i = 0;
    #while( i < 3 )
      #if( i = j )
        #declare B[ i ][ j ] = 1;
      #else
        #declare B[ i ][ j ] = 0;
      #end 
      
      #declare I[ i ][ j ] = A[ i ][ j ];               
                    
      #local i = i + 1;
    #end
    
    #local j = j + 1;
  #end
       
  Debug_Message( "main loop over the columns to be reduced\n" )     
  // main loop over the columns to be reduced
  #local i = 0;
  #while( i < 3 )
    #local big = 0;
                
    Debug_Message( "outer loop over search for a pivot element\n" )   
    // outer loop over search for a pivot element
    #local j = 0;
    #while( j < 3 )
      #if( ipiv[ j ] != 1 )
        #local k = 0;
        #while( k < 3 )
          #if( ipiv[ k ] = 0 )
            #if( abs( I[ j ][ k ] ) >= big )
               #local big = abs( I[ j ][ k ] );
               #local irow = j;
               #local icol = k;
            #end
          #end
          
          #local k = k + 1;
        #end
      #end
      
      #local j = j + 1;
    #end
    
    #local ipiv[ icol ] = ipiv[ icol ] + 1;
    
    Debug_Message( "now we have the pivot element, interchange rows, if needed, using book-keeping\n" )   
    // now we have the pivot element, interchange rows, if needed, using book-keeping
    #if( irow != icol )
      #local l = 0;
      #while( l < 3 )
        swap( I[ irow ][ l ], I[ icol ][ l ] )
        swap( B[ irow ][ l ], B[ icol ][ l ] )
        
        #local l = l + 1;
      #end
    #end 
       
    Debug_Message( "divide the pivot row by the pivot element\n" )  
    // divide the pivot row by the pivot element
    #local indxr[ i ] = irow;
    #local indxc[ i ] = icol;
    #if( I[ icol ][ icol ] = 0 )
      #error "GaussJordan4x4: Singular Matrix\n"
    #end 
    #local pivinv = 1 / I[ icol ][ icol ];
    #declare I[ icol ][ icol ] = 1;   
    
    #local l = 0;
    #while( l < 3 ) 
      #declare I[ icol ][ l ] = I[ icol ][ l ] * pivinv;
      #declare B[ icol ][ l ] = B[ icol ][ l ] * pivinv;
      
      #local l = l + 1;
    #end  
     
    Debug_Message( "reduce the rows, except for the pivot one\n" )  
    // reduce the rows, except for the pivot one
    #local ll = 0;
    #while( ll < 3 )   
      #if( ll != icol )
        #local dum = I[ ll ][ icol ];    
        #declare I[ ll ][ icol ] = 0;
        #local l = 0;
        #while( l < 3 )     
          #declare I[ ll ][ l ] = I[ ll ][ l ] - I[ icol ][ l ] * dum;
          #declare B[ ll ][ l ] = B[ ll ][ l ] - B[ icol ][ l ] * dum;
          
          #declare l = l + 1;
        #end
      #end 
      
      #local ll = ll + 1;
    #end
    
    #local i = i + 1;
  #end 
       
  Debug_Message( "finished the main loop over columns of the reduction\n" )  
  Debug_Message( "need to unscramble the rows that we switched around\n" )  
  // finished the main loop over columns of the reduction
  // need to unscramble the rows that we switched around
  #local l = 2;
  #while( l >= 0 ) 
    #if( indxr[ l ] != indxc[ l ] )
      #local k = 0;
      #while( k < 3 ) 
        swap( I[ k ][ indxr[ l ] ], I[ k ][ indxc[ l ] ] )
        
        #local k = k + 1;
      #end
    #end   
    
    #local l = l - 1;
  #end 
  
  Debug_Message( "Leaving GaussJordan4x4\n" )
  
  // and we are done
#end 

#macro MatrixMultiply3x3( A, B, M )
  Debug_Message( "Entering MatrixMultiply4x4\n" )
 
  #local r = 0;
  #while( r < 3 )
    #local c = 0;
    #while( c < 3 ) 
      #declare M[ r ][ c ] = 0;
    
      #local i = 0;
      #while( i < 3 ) 
        #declare M[ r ][ c ] = M[ r ][ c ] + A[ r ][ i ] * B[ i ][ c ];               
                       
        #local i = i + 1;
      #end
      
      #local c = c + 1;
    #end
    
    #local r = r + 1;
  #end    
  
  Debug_Message( "Leaving MatrixMultiply4x4\n" )
#end

#macro FindAffine( p0, p1, p2, p3, q0, q1, q2, q3, M )    
  Debug_Message( "Entering FindAffine\n" )

  #declare p3 = ( p0 + vcross( p2 - p1, p0 - p1 ) / sqrt( vlength( vcross( p2 - p1, p0 - p1 ) ) ) );
  #declare q3 = ( q0 + vcross( q2 - q1, q0 - q1 ) / sqrt( vlength( vcross( q2 - q1, q0 - q1 ) ) ) );
  
  #local Tp = array[ 3 ][ 3 ] { { p1.x - p0.x, p2.x - p0.x, p3.x - p0.x },
                                { p1.y - p0.y, p2.y - p0.y, p3.y - p0.y },
                                { p1.z - p0.z, p2.z - p0.z, p3.z - p0.z } };
  #local Tq = array[ 3 ][ 3 ] { { q1.x - q0.x, q2.x - q0.x, q3.x - q0.x },
                                { q1.y - q0.y, q2.y - q0.y, q3.y - q0.y },
                                { q1.z - q0.z, q2.z - q0.z, q3.z - q0.z } }; 
                                
  #local Tp_inv = array[ 3 ][ 3 ]; 
  #local sol = array[ 3 ][ 3 ];
  #local M_inv = array[ 3 ][ 3 ];
  
  GaussJordan3x3( Tp, sol, Tp_inv ) 
  MatrixMultiply3x3( Tq, Tp_inv, M )
  
  Debug_Message( "Leaving FindAffine\n" )
#end

 

#version FINDAFFINE_INC_TEMP;
#end